/// Header files
#include <windows.h>
#include "gpslib.h" 

/// Macros
#define MINIMUM_BAUD     300
#define MAXIMUM_BAUD     921600
#define RECEIVE_BUF_MAX  350
#define DEGREE           176     ///< ASCII value for degree sign
#define DOUBLE_QUOTE     34      ///< ASCII value for Double quote sign (")
#define SINGLE_QUOTE     39      ///< ASCII value for single quote sign (')
#define ASTERISK         42      ///< ASCII value for star sign (*)
#define HEX_BASE         16      ///< Hexadecimal base = 16

//******************************************************************************
/// Configures COM port for specified baud rate.
/// @param[in]    port         COM port handle.
/// @param[in]    portNumber   COM port number.
/// @param[in]    baudRate     Baud rate for communication.
/// @retval       TRUE         Success.
///               FALSE        Failure.
BOOL PortOpen(HANDLE *port, wchar_t *comPortNumber, DWORD baudRate)
{
    DCB portDCB;                                              ///< COM port configuration structure
    wchar_t fileName[8] = L"COM";
    BOOL returnValue = FALSE;
    COMMTIMEOUTS comTimeOut;

    wcscat_s(fileName, sizeof(fileName), comPortNumber);      ///< fileName = COMx
    wcscat_s(fileName, sizeof(fileName), L":");               ///< fileName = COMx:

    /// Open interface to reader.
    *port = CreateFile(fileName, GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, FILE_FLAG_WRITE_THROUGH, NULL);
    if (*port == INVALID_HANDLE_VALUE)
    {
        printf("Error Opening COM Port\n");
        return FALSE;
    }
 
    /// COM Port Configuration.
    portDCB.DCBlength = sizeof (DCB);                         ///< Initialize the DCBlength member.
    GetCommState (*port, &portDCB);                           ///< Get the default port setting information.
    /// Change the DCB structure settings.
    portDCB.BaudRate = baudRate;                              ///< Current baud.
    portDCB.fBinary = TRUE;                                   ///< Binary mode; no EOF check. 
    portDCB.fParity = FALSE;                                  ///< Disable parity checking. 
    portDCB.fOutxCtsFlow = FALSE;                             ///< No CTS output flow control.
    portDCB.fOutxDsrFlow = FALSE;                             ///< No DSR output flow control. 
    portDCB.fDtrControl = DTR_CONTROL_DISABLE;                ///< Disable DTR flow control type. 
    portDCB.fDsrSensitivity = FALSE;                          ///< DSR sensitivity. 
    portDCB.fTXContinueOnXoff = TRUE;                         ///< XOFF continues Tx.
    portDCB.fOutX = FALSE;                                    ///< No XON/XOFF out flow control. 
    portDCB.fInX = FALSE;                                     ///< No XON/XOFF in flow control. 
    portDCB.fErrorChar = FALSE;                               ///< Disable error replacement. 
    portDCB.fNull = FALSE;                                    ///< Disable null stripping. 
    portDCB.fRtsControl = RTS_CONTROL_DISABLE;                ///< Disable RTS flow control. 
    portDCB.fAbortOnError = FALSE;                            ///< Do not abort reads/writes on error.
    portDCB.ByteSize = 8;                                     ///< Number of bits/byte, 4-8. 
    portDCB.Parity = NOPARITY;                                ///< 0-4 = no, odd, even, mark, space. 
    portDCB.StopBits = ONESTOPBIT;                            ///< 0, 1, 2 = 1, 1.5, 2.
 
    /// Configure the port according to the specifications of the DCB structure.
    if (!SetCommState (*port, &portDCB))
    {
      printf("Error Configuring COM Port\n");                 ///< Could not configure the serial port.      
      return FALSE;
    }
 
    /// PurgeComm(*port, PURGE_RXCLEAR);                      ///< Clear receiver buffer.
    /// Get communication time out values
    returnValue = GetCommTimeouts(*port, &comTimeOut);
    comTimeOut.ReadIntervalTimeout = 10;
    comTimeOut.ReadTotalTimeoutMultiplier = 1;
    comTimeOut.ReadTotalTimeoutConstant = 1;
    /// Set communication time out values
    SetCommTimeouts(*port, &comTimeOut);
    return TRUE;
}

//******************************************************************************
/// Close serial port
/// @param[in]    port     COM port.
/// @retval       TRUE     Success.
///               FALSE    Failure.
BOOL PortClose(HANDLE *port)
{
    if (*port == NULL)
    {
        return FALSE;
    }
    CloseHandle(*port);                                         ///< Close port.
    *port = NULL;
    return TRUE;
}

//******************************************************************************
int wmain(void)
{
    HANDLE portHandle = NULL;
    DWORD noOfBytesRead = 0;
    DWORD baudRate = 0;
    BOOL returnValue = FALSE;
    BOOL loopTerminate = FALSE;
    DWORD maxRead = 5;
    CHAR *start = NULL;
    CHAR *end = NULL;
    CHAR dataId[7] = "$GPGGA";
    CHAR receiveBuffer[RECEIVE_BUF_MAX + 1] = {0};
    wchar_t comPortNumber[4] = {0};
    DWORD selectDatatype = 0;
    errorType error = ERROR_FREE;
    gpggaParameter ggaStructInfo = {0};
    gpgllParameter gllStructInfo = {0};
    gprmcParameter rmcStructInfo = {0};
    
    /// Get COM port connected to GPS device.
    printf("Select COM port to open (example = 1, 2, 3...) ");
    wscanf_s(L"%s", comPortNumber);

    /// Set COM port baud rate.
    while(loopTerminate == FALSE)
    {
        printf("Enter Baud rate (bps) = ");
        scanf_s("%u", &baudRate);

        if(baudRate < MINIMUM_BAUD || baudRate > MAXIMUM_BAUD)
        {
            printf("Entered baud rate is not supported by this device.\n");
        }
        else
        {
            loopTerminate = TRUE;
        }
    }

    /// Open selected COM port
    returnValue = PortOpen(&portHandle, comPortNumber, baudRate);               ///< Open serial port.
    if (returnValue == FALSE)
    {
        printf("Could not open COM port");
        printf("\nPress enter to exit.");
        fflush(stdin);                                                          ///< Clears standard input output buffer.
        getchar();
        return FALSE;
    }   
   
    loopTerminate = FALSE;
    while(loopTerminate == FALSE)
    {
        printf("\nPress 1 for GPGGA data type.\nPress 2 for GPGLL data type.\nPress 3 for GPRMC data type.");
        scanf_s("%u", &selectDatatype);

        selectDatatype -= 1;
        if(selectDatatype != GPGGA && selectDatatype != GPGLL && selectDatatype != GPRMC)
        {
            printf("Entered data type is not supported.\n");
        }
        else
        {
            loopTerminate = TRUE;
        }
    }

    /// Integer Data type to string data type conversion.
    switch(selectDatatype)
    {
    case GPGGA:
        {
            strcpy_s(dataId, sizeof(dataId), "$GPGGA");
            break;
        }
    case GPGLL:
        {
            strcpy_s(dataId, sizeof(dataId), "$GPGLL");
            break;
        }
    case GPRMC:
        {
            strcpy_s(dataId, sizeof(dataId), "$GPRMC");
        }
    }

    /// Scan COM port for desired data type.
    while(maxRead > 0)
    {
        maxRead--;
        memset(receiveBuffer, 0, sizeof(receiveBuffer));
        returnValue = ReadFile(portHandle, (CHAR*)receiveBuffer, RECEIVE_BUF_MAX ,&noOfBytesRead, NULL);        ///< Read serial port.
        if(returnValue == FALSE)                                                                                ///< if read fails.
        {
            printf("\nCOM port read failed.");
            printf("\nPress enter to exit.");
            fflush(stdin);                                                          ///< Clears standard input output buffer.
            getchar();
            return FALSE;
        }

        start = strstr(receiveBuffer, dataId);
        /// If data id is not found.
        if(start == NULL)
        {
            error = DATATYPE_NOT_FOUND;
            continue;
        }
        
        end = strchr(start, ASTERISK);
        /// Checks existence of * and checksum for successful completion of the string.
        if(end == NULL || *(end + 1) == FALSE || *(end + 2) == FALSE)                                               
        {
            error = INVALID_DATA;
            continue;
        }
        error = ERROR_FREE;
        break;
    }
    if(error!= ERROR_FREE)
    {
        printf("\nRead failed with error code = %u", error);
        printf("\nPress enter to exit.");
        fflush(stdin);                                                              ///< Clears standard input output buffer.
        getchar();
        return FALSE;
    }
    
    switch(selectDatatype)
    {
    case GPGGA:
        {
            /// Extract all fields individually from received data.
            returnValue = ExtractAllFieldsFromGpsData(start, selectDatatype, &ggaStructInfo, &error);
            if(returnValue == FALSE)
            {
                printf("\nProcessing failed with error code = %u", error);
                printf("\nPress enter to exit.");
                fflush(stdin);
                getchar();
                return FALSE;
            }

            printf("\nId processed = %s", "GPGGA");                                 ///< Id processed.
            /// UTC time information.
            printf("\nUTC Time (HH: MM: SS.SSSSSS) = %02u : %02u : %09.6f", ggaStructInfo.currentTimeInfo.utcHour, 
                                                                            ggaStructInfo.currentTimeInfo.utcMin,
                                                                            ggaStructInfo.currentTimeInfo.utcSec);
            /// Latitude information.
            printf("\nLatitude = %02u%c %02u%c %07.4f%c %c", ggaStructInfo.latitudeInfo.latDeg, DEGREE,
                                                             ggaStructInfo.latitudeInfo.latMin, SINGLE_QUOTE,
                                                             ggaStructInfo.latitudeInfo.latSec, DOUBLE_QUOTE,
                                                             ggaStructInfo.latitudeInfo.latDirection);
            /// Longitude information.
            printf("\nLongitude = %03u%c %02u%c %07.4f%c %c", ggaStructInfo.longitudeInfo.longDeg, DEGREE,
                                                              ggaStructInfo.longitudeInfo.longMin, SINGLE_QUOTE,
                                                              ggaStructInfo.longitudeInfo.longSec, DOUBLE_QUOTE,
                                                              ggaStructInfo.longitudeInfo.longDirection);
            /// Quality of GPS signal recorded.
            switch(ggaStructInfo.positionInfo)
            {
            case FAILURE:
                {
                    printf("\nQuality of signal (Position): Failure.");
                    break;
                }
            case VALID_SPS:
                {
                    printf("\nQuality of signal (Position): SPS fix.");
                    break;
                }
            case VALID_DGPS:
                {
                    printf("\nQuality of signal (Position): DGPS fix.");
                    break;
                }
            case VALID_PPS:
                {
                    printf("\nQuality of signal (Position): PPS fix.");
                    break;
                }
            case REAL_TIME_KINEMATIC:
                {
                    printf("\nQuality of signal (Position): Real time.");
                    break;
                }
            case FLOAT_RTK:
                {
                    printf("\nQuality of signal (Position): Float RTK.");
                    break;
                }
            case ESTIMATED:
                {
                    printf("\nQuality of signal (Position): Estimated.");
                    break;
                }
            case MANUAL_INPUT_MODE:
                {
                    printf("\nQuality of signal (Position): Manual mode.");
                    break;
                }
            case SIMULATION_MODE:
                {
                    printf("\nQuality of signal (Position): Simulation mode.");
                    break;
                }
            default:
                {
                    printf("\nQuality of signal (Position): Invalid.");
                }
            }

            printf("\nSatellites in use = %02u", ggaStructInfo.satelliteUsed);      ///< Number of satellites in use.
            printf("\nHDOP value = %f", ggaStructInfo.hdopInfo);                    ///< Horizontal dilution of precision (HDOP).
            printf("\nAltitude = %f %c", ggaStructInfo.altitudeInfo.height, ggaStructInfo.altitudeInfo.altiUnit);   ///< Altitude information.
            printf("\nGeoid = %f %c", ggaStructInfo.geoidInfo.height, ggaStructInfo.geoidInfo.geoidUnit);           ///< Geoid information.
            printf("\nTime DGPS update value = %f Sec", ggaStructInfo.timeDgps);    ///< Time difference in seconds (Valid for position type- DGPS).
            printf("\nDGPS station Id = %u", ggaStructInfo.stationIdDgps);          ///< DGPS station id (Valid for position type- DGPS).
            printf("\nChecksum = 0x%X", ggaStructInfo.checksumData);                ///< Checksum.      
            /// Checksum verification.
            if(ggaStructInfo.checkSumVerify == 'Y')
            {
                printf(" (Verification- OK)");
            }
            else
            {
                printf(" (Verification- Not OK)");
            }
            break;
        }
    case GPGLL:
        {
            /// Extract all fields individually from received data.
            returnValue = ExtractAllFieldsFromGpsData(start, selectDatatype, &gllStructInfo, &error);
            if(returnValue == FALSE)
            {
                printf("\nProcessing failed with error code = %u", error);
                printf("\nPress enter to exit.");
                fflush(stdin);
                getchar();
                return FALSE;
            }

            printf("\nId processed = %s", "GPGLL");                                 ///< Id processed.
            /// Latitude information.
            printf("\nLatitude = %02u%c %02u%c %07.4f%c %c", gllStructInfo.latitudeInfo.latDeg, DEGREE,
                                                             gllStructInfo.latitudeInfo.latMin, SINGLE_QUOTE,
                                                             gllStructInfo.latitudeInfo.latSec, DOUBLE_QUOTE,
                                                             gllStructInfo.latitudeInfo.latDirection);
            /// Longitude information.
            printf("\nLongitude = %03u%c %02u%c %07.4f%c %c", gllStructInfo.longitudeInfo.longDeg, DEGREE,
                                                              gllStructInfo.longitudeInfo.longMin, SINGLE_QUOTE,
                                                              gllStructInfo.longitudeInfo.longSec, DOUBLE_QUOTE,
                                                              gllStructInfo.longitudeInfo.longDirection);
            /// UTC time information.
            printf("\nUTC Time (HH: MM: SS.SSSSSS) = %02u : %02u : %09.6f", gllStructInfo.currentTimeInfo.utcHour, 
                                                                            gllStructInfo.currentTimeInfo.utcMin,
                                                                            gllStructInfo.currentTimeInfo.utcSec);
            /// Validation status of received data.
            switch(gllStructInfo.statusOfData)
            {
            case VALID:
                {
                    printf("\nReceived data status: Valid");
                    break;
                }
            case INVALID:
                {
                    printf("\nReceived data status: Invalid");
                    break;
                }
            default:
                {
                    printf("\nReceived data status: UNDEFINED");
                    break;
                }
            }

            /// Quality of GPS signal recorded.
            switch(gllStructInfo.positionInfo)
            {
            case 'N':
                {
                    printf("\nQuality of signal (Position): N (Failure).");
                    break;
                }
            case 'A':
                {
                    printf("\nQuality of signal (Position): A (Fix).");
                    break;
                }
            case 'E':
                {
                    printf("\nQuality of signal (Position): E (Estimated).");
                    break;
                }
            case 'M':
                {
                    printf("\nQuality of signal (Position): M (Manual mode).");
                    break;
                }
            case 'S':
                {
                    printf("\nQuality of signal (Position): S (Simulation mode).");
                    break;
                }
            default:
                {
                    printf("\nQuality of signal (Position): Invalid.");
                }
            }

            printf("\nChecksum = 0x%X", gllStructInfo.checksumData);                ///< Checksum.
            /// Checksum verification.
            if(gllStructInfo.checkSumVerify == 'Y')
            {
                printf(" (Verification- OK)");
            }
            else
            {
                printf(" (Verification- Not OK)");
            }
            break;
        }
    case GPRMC:
        {
            /// Extract all fields individually from received data.
            returnValue = ExtractAllFieldsFromGpsData(start, selectDatatype, &rmcStructInfo, &error);
            if(returnValue == FALSE)
            {
                printf("\nProcessing failed with error code = %u", error);
                printf("\nPress enter to exit.");
                fflush(stdin);
                getchar();
                return FALSE;
            }

            printf("\nId processed = %s", "GPRMC");                                 ///< Id processed.
            /// UTC time information.
            printf("\nUTC Time (HH: MM: SS.SSSSSS) = %02u : %02u : %09.6f", rmcStructInfo.currentTimeInfo.utcHour, 
                                                                            rmcStructInfo.currentTimeInfo.utcMin,
                                                                            rmcStructInfo.currentTimeInfo.utcSec);
            /// Validation status of received data.
            switch(rmcStructInfo.statusOfData)
            {
            case VALID:
                {
                    printf("\nReceived data status: Valid");
                    break;
                }
            case INVALID:
                {
                    printf("\nReceived data status: Invalid");
                    break;
                }
            default:
                {
                    printf("\nReceived data status: UNDEFINED");
                    break;
                }
            }

            ///< Latitude information.
            printf("\nLatitude = %02u%c %02u%c %07.4f%c %c", rmcStructInfo.latitudeInfo.latDeg, DEGREE,
                                                             rmcStructInfo.latitudeInfo.latMin, SINGLE_QUOTE,
                                                             rmcStructInfo.latitudeInfo.latSec, DOUBLE_QUOTE,
                                                             rmcStructInfo.latitudeInfo.latDirection);
            ///< Longitude information.
            printf("\nLongitude = %03u%c %02u%c %07.4f%c %c", rmcStructInfo.longitudeInfo.longDeg, DEGREE,
                                                              rmcStructInfo.longitudeInfo.longMin, SINGLE_QUOTE,
                                                              rmcStructInfo.longitudeInfo.longSec, DOUBLE_QUOTE,
                                                              rmcStructInfo.longitudeInfo.longDirection);
            
            printf("\nSpeed = %f KMPH", rmcStructInfo.speedInKmph);                             ///< Speed in KMPH.     
            printf("\nTrue Course over ground = %f%c", rmcStructInfo.trueCourse, DEGREE);       ///< Course over ground.
            /// UTC date information.
            printf("\nDate (DD/ MM/ YY) = %02u/ %02u/ %02u", rmcStructInfo.currentDateInfo.utcDate,
                                                             rmcStructInfo.currentDateInfo.utcMonth,
                                                             rmcStructInfo.currentDateInfo.utcYear);    
            ///< Magnetic variation information.
            printf("\nMagnetic Variation = %f%c %c", rmcStructInfo.magneticData.magVarDeg, DEGREE, 
                                                     rmcStructInfo.magneticData.magVarDirection);
            /// Quality of GPS signal recorded.
            switch(rmcStructInfo.positionInfo)
            {
            case 'N':
                {
                    printf("\nQuality of signal (Position): N (Failure).");
                    break;
                }
            case 'A':
                {
                    printf("\nQuality of signal (Position): A (Fix).");
                    break;
                }
            case 'E':
                {
                    printf("\nQuality of signal (Position): E (Estimated).");
                    break;
                }
            case 'M':
                {
                    printf("\nQuality of signal (Position): M (Manual mode).");
                    break;
                }
            case 'S':
                {
                    printf("\nQuality of signal (Position): S (Simulation mode).");
                    break;
                }
            default:
                {
                    printf("\nQuality of signal (Position): Invalid.");
                }
            }

            printf("\nChecksum = 0x%X", rmcStructInfo.checksumData);                ///< Checksum.
            /// Checksum verification.
            if(rmcStructInfo.checkSumVerify == 'Y')
            {
                printf(" (Verification- OK)");
            }
            else
            {
                printf(" (Verification- Not OK)");
            }
            break;
        }
    }
    printf("\n\nPress enter to exit.");
    fflush(stdin);                                                                  ///< Clears standard input output buffer.
    getchar();
    return TRUE;
}