/// @file			gpslib.h
/// @copyright  	Copyright (c) 2014 Toradex AG.
/// $Author:		khv
/// $Date: 			$ 2014-06-04 $
/// @brief:			File to support GPS interfacing.  

/// Header files
#include <windows.h>
#include "gpslib.h"

/// Macros
#define COMMA_LENGTH 1
#define ASTERISK_LENGTH 1

BOOL ConvertAsciiToHex(CHAR *source, DWORD *hexResult)
{
	DWORD hexArray[8] = {0};
	DWORD count = 0;
	DWORD shiftLeft = 0;

	if(source == NULL || hexResult == NULL)
	{
		return FALSE;
	}
	
	while(count < 8)
	{
		if(*source >= '0' && *source <= '9')
		{
			hexArray[count] = *source - '0';
			count++;
			source++;
		}
		else if(*source >= 'A' && *source <= 'F')
		{
			hexArray[count] = *source - 'A' + 10;					///< Hex 'A' = 10
			count++;
			source++;
		}
		else if(*source >= 'a' && *source <= 'f')
		{
			hexArray[count] = *source - 'a' + 10;					///< Hex 'a' = 10
			count++;
			source++;
		}
		else
		{
			break;
		}
	}
	while(count)
	{
		count--;
		*hexResult = *hexResult + (hexArray[count] << shiftLeft);
		shiftLeft += BITS_IN_NIBBLE;
	}
	return TRUE;
}

BOOL GenerateCheckSum(CHAR *source, BYTE* checkSum)
{
	DWORD index = 0;
	if(source == NULL || checkSum == NULL)
	{
		return FALSE;
	}

	if(*(source + index) == '$')									
	{
		index = 1;													///< Ignoring $.						
	}
	
	while(*(source + index) != '*' && *(source + index) != '\0')
	{
		*checkSum ^= *(source + index);								///< X-OR operation.
		index++;
	}
	return TRUE;
}

BOOL ExtractSubString(CHAR *SourceString, CHAR *destinationString, DWORD destinationSize, CHAR terminator, BOOL skipTerminatorOnce)
{
	DWORD sourceIndex = 0;
	DWORD destinationIndex = 0;
	BOOL loopTerminate = FALSE;
	DWORD terminatorSkipped = 0;
	char terminatorBackup = terminator;

	if(SourceString == NULL || destinationString == NULL || (skipTerminatorOnce != TRUE && skipTerminatorOnce != FALSE))
	{
		return FALSE;
	}
	
	memset(destinationString, 0, destinationSize);
	while(loopTerminate == FALSE)
	{
		while(*(SourceString + sourceIndex) != terminator && *(SourceString + sourceIndex) != '\0')
		{
			terminator = terminatorBackup;
			if(destinationIndex >= destinationSize - 1)
			{
				break;
			}

			*(destinationString + destinationIndex) = *(SourceString + sourceIndex);
			destinationIndex++;
			sourceIndex++;
		}
		if(skipTerminatorOnce == FALSE)
		{
			loopTerminate = TRUE;
		}
		else
		{
			terminatorSkipped = 1;
			skipTerminatorOnce = FALSE;
			terminator = '\0';
		}
	}
	return TRUE;
}

BOOL GetUtcTime(CHAR *timeSource, timeParameter *timeStore)
{
	CHAR destination[10] = {0};
	DWORD destinationIndex = 0;
	DWORD sourceIndex = 0;
	DWORD hourMinSec = 0;
	FLOAT secondFraction = 0.0;

	if(timeSource == NULL || timeStore == NULL)
	{
		return FALSE;
	}

	while(timeSource[sourceIndex] != '.' && timeSource[sourceIndex] != '\0')
	{
		destination[destinationIndex] = timeSource[sourceIndex];
		sourceIndex++;
		destinationIndex++;
	}
	hourMinSec = atoi(destination);										///< ASCII to integer conversion.
	memset(destination, 0, sizeof(destination));
	destinationIndex = 0;

	if(timeSource[sourceIndex] == '.')
	{
		while(timeSource[sourceIndex] != '\0')
		{
			destination[destinationIndex] = timeSource[sourceIndex];
			sourceIndex++;
			destinationIndex++;
		}
	}
	secondFraction = (FLOAT)atof(destination);							///< ASCII to float conversion.
	timeStore->utcSec = (FLOAT)(hourMinSec % 100) + secondFraction;
	hourMinSec /= 100;
	timeStore->utcMin = hourMinSec % 100;
	hourMinSec /= 100;
	timeStore->utcHour = hourMinSec % 100;
	return TRUE;
}

BOOL GetUtcDate(CHAR *dateSource, dateParameter *dateStore)
{
	CHAR destination[6] = {0};
	DWORD index = 0;
	DWORD dateMonthYear = 0;

	if(dateSource == NULL || dateStore == NULL)
	{
		return FALSE;
	}

	while(dateSource[index] != '\0')
	{
		destination[index] = dateSource[index];
		index++;
	}
	dateMonthYear = atoi(destination);									///< ASCII to integer conversion.
	dateStore->utcYear = dateMonthYear % 100;
	dateMonthYear /= 100;
	dateStore->utcMonth = dateMonthYear % 100;
	dateMonthYear /= 100;
	dateStore->utcDate = dateMonthYear % 100;
	return TRUE;
}

BOOL GetLatitude(CHAR *latitudeSource, latitudeParameter *latitudeStore)
{
	CHAR destination[10] = {0};
	DWORD sourceIndex = 0;
	DWORD destinationIndex = 0;
	DWORD degreeMinute;

	if(latitudeSource == NULL || latitudeStore == NULL)
	{
		return FALSE;
	}

	while(latitudeSource[sourceIndex] != '.' && latitudeSource[sourceIndex] != ',' && latitudeSource[sourceIndex] != '\0')
	{
		destination[destinationIndex] = latitudeSource[sourceIndex];
		destinationIndex++;
		sourceIndex++;
	}
	degreeMinute = atoi(destination);									///< ASCII to integer conversion.
	memset(destination, 0, sizeof(destination));
	destinationIndex = 0;

	if(latitudeSource[sourceIndex] == '.')
	{
		while(latitudeSource[sourceIndex] != ',' && latitudeSource[sourceIndex] != '\0')
		{
			destination[destinationIndex] = latitudeSource[sourceIndex];
			sourceIndex++;
			destinationIndex++;
		}
		sourceIndex++;
	}
	latitudeStore->latSec = (FLOAT)atof(destination) * 60;				///< ASCII to float conversion.
	latitudeStore->latMin = degreeMinute % 100;
	degreeMinute /= 100;
	latitudeStore->latDeg = degreeMinute % 100;

	if(latitudeSource[sourceIndex] == '\0' || latitudeSource[sourceIndex] == ',')
	{
		latitudeStore->latDirection = '\0';
	}
	else
	{
		latitudeStore->latDirection = latitudeSource[sourceIndex];		///< Copy direction parameter.
	}
	return TRUE;
}

BOOL GetLongitude(CHAR *longitudeSource, longitudeParameter *longitudeStore)
{
	CHAR destination[10] = {0};
	DWORD sourceIndex = 0;
	DWORD destinationIndex = 0;
	DWORD degreeMinute = 0;

	if(longitudeSource == NULL || longitudeStore == NULL)
	{
		return FALSE;
	}

	while(longitudeSource[sourceIndex] != '.' && longitudeSource[sourceIndex] != ',' && longitudeSource[sourceIndex] != '\0')
	{
		destination[destinationIndex] = longitudeSource[sourceIndex];
		destinationIndex++;
		sourceIndex++;
	}
	degreeMinute = atoi(destination);									///< ASCII to integer conversion.
	memset(destination, 0, sizeof(destination));
	destinationIndex = 0;

	if(longitudeSource[sourceIndex] == '.')
	{
		while(longitudeSource[sourceIndex] != ',' && longitudeSource[sourceIndex] != '\0')
		{
			destination[destinationIndex] = longitudeSource[sourceIndex];
			sourceIndex++;
			destinationIndex++;
		}
		sourceIndex++;
	}
	longitudeStore->longSec = (FLOAT)atof(destination) * 60;				///< ASCII to float conversion.
	longitudeStore->longMin = degreeMinute % 100;
	degreeMinute /= 100;
	longitudeStore->longDeg = (WORD)degreeMinute % 1000;

	if(longitudeSource[sourceIndex] == '\0' ||longitudeSource[sourceIndex] == ',')
	{
		longitudeStore->longDirection = '\0';
	}
	else
	{
		longitudeStore->longDirection = longitudeSource[sourceIndex];		///< Copy direction parameter.
	}
	return TRUE;
}

BOOL GetAltitude(CHAR *altitudeSource, altitudeParameter *altitudeStore)
{
	CHAR destination[10] = {0};
	DWORD sourceIndex = 0;
	DWORD destinationIndex = 0;

	if(altitudeSource == NULL || altitudeStore == NULL)
	{
		return FALSE;
	}
		
	while(altitudeSource[sourceIndex] != ',' && altitudeSource[sourceIndex] != '\0')
	{
		destination[destinationIndex] = altitudeSource[sourceIndex];
		sourceIndex++;
		destinationIndex++;
	}
	altitudeStore->height = (FLOAT)atof(destination);						///< ASCII to float conversion.
	sourceIndex++;

	if(altitudeSource[sourceIndex] == '\0')
	{
		altitudeStore->altiUnit = '\0';
	}
	else
	{
		altitudeStore->altiUnit = altitudeSource[sourceIndex];				///< Copy Altitude unit.
	}
	return TRUE;
}

BOOL GetGeoid(CHAR *geoidSource, geoidParameter *geoidStore)
{
	CHAR destination[10] = {0};
	DWORD sourceIndex = 0;
	DWORD destinationIndex = 0;

	if(geoidSource == NULL || geoidStore == NULL)
	{
		return FALSE;
	}
		
	while(geoidSource[sourceIndex] != ',' && geoidSource[sourceIndex] != '\0')
	{
		destination[destinationIndex] = geoidSource[sourceIndex];
		sourceIndex++;
		destinationIndex++;
	}
	geoidStore->height = (FLOAT)atof(destination);							///< ASCII to float conversion.
	sourceIndex++;

	if(geoidSource[sourceIndex] == '\0')
	{
		geoidStore->geoidUnit = '\0';
	}
	else
	{
		geoidStore->geoidUnit = geoidSource[sourceIndex];					///< Copy Geoid unit.
	}
	return TRUE;
}

BOOL GetMagneticVariation(CHAR *magnetSource, magnetParameter *magnetStore)
{
	CHAR destination[10] = {0};
	DWORD sourceIndex = 0;
	DWORD destinationIndex = 0;

	if(magnetSource == NULL || magnetStore == NULL)
	{
		return FALSE;
	}
		
	while(magnetSource[sourceIndex] != ',' && magnetSource[sourceIndex] != '\0')
	{
		destination[destinationIndex] = magnetSource[sourceIndex];
		destinationIndex++;
		sourceIndex++;
	}
	magnetStore->magVarDeg = (FLOAT)atof(destination);						///< ASCII to float conversion.
	sourceIndex++;

	if(magnetSource[sourceIndex] == '\0')
	{
		magnetStore->magVarDirection = '\0';
	}
	else
	{
		magnetStore->magVarDirection = magnetSource[sourceIndex];			///< Copy direction parameter.
	}
	return TRUE;
}

BOOL ExtractAllFieldsFromGpsData(CHAR *sourceString, gpsDataType dataId, VOID *ptrStoreResults, errorType *error)
{
	gpggaParameter *ggaInfoStruct = NULL;
	gpgllParameter *gllInfoStruct = NULL;
	gprmcParameter *rmcInfoStruct = NULL;
	CHAR extractData[15] = {0};
	DWORD index = 0;
	DWORD totalFieldsInSrc = 0;
	DWORD sourceIndex = 0;
	BOOL retStatus = FALSE;
	BYTE checkSum = 0;

	*error = ERROR_FREE;
	if(!((dataId == GPGGA) || (dataId == GPGLL) || (dataId == GPRMC)))
	{
		*error = INVALID_DATATYPE;
		return FALSE;
	}

	if(*sourceString == '$')
	{
		index = 7;										///< "$GPxxx," = Total 7 characters.					
	}
	else if(*sourceString == 'G')
	{
		index = 6;										///< "GPxxx," = Total 6 characters.
	}
	else
	{
		*error = INVALID_DATA;
		return FALSE;
	}

	while(*(sourceString + sourceIndex) != '\0' && *(sourceString + sourceIndex) != '*')
	{
		if(*(sourceString + sourceIndex) == ',')
		{
			totalFieldsInSrc++;
		}
		sourceIndex++;
	}

	if((dataId == GPGGA && GPGGA_TOTAL_FIELDS != totalFieldsInSrc) ||
	   (dataId == GPGLL && GPGLL_TOTAL_FIELDS != totalFieldsInSrc) ||
	   (dataId == GPRMC && GPRMC_TOTAL_FIELDS != totalFieldsInSrc) ||
	   (sourceIndex > MAX_GPS_DATA_STRING_LENGTH))
	{
		*error = INVALID_DATA;
		return FALSE;
	}

	switch(dataId)
	{
	case GPGGA:
		{
			ggaInfoStruct = (gpggaParameter*)ptrStoreResults;

			ExtractSubString(sourceString + index, extractData, sizeof(extractData), ',', FALSE);				///< Extracts UTC time.
			GetUtcTime(extractData, &(ggaInfoStruct->currentTimeInfo));											///< Stores UTC time.
			
			index += strlen(extractData) + COMMA_LENGTH;
			ExtractSubString(sourceString + index, extractData, sizeof(extractData), ',', TRUE);				///< Extracts latitude.
			GetLatitude(extractData, &(ggaInfoStruct->latitudeInfo));											///< Stores latitude.

			index += strlen(extractData) + COMMA_LENGTH;
			ExtractSubString(sourceString + index, extractData, sizeof(extractData), ',', TRUE);				///< Extracts longitude.
			GetLongitude(extractData, &(ggaInfoStruct->longitudeInfo));											///< Stores longitude.

			index += strlen(extractData) + COMMA_LENGTH;
			ExtractSubString(sourceString + index, extractData, sizeof(extractData), ',', FALSE);				///< Extracts position type.
			ggaInfoStruct->positionInfo = atoi(extractData);													///< ASCII to integer conversion.

			index += strlen(extractData) + COMMA_LENGTH;
			ExtractSubString(sourceString + index, extractData, sizeof(extractData), ',', FALSE);				///< Extracts number of satellites in use.
			ggaInfoStruct->satelliteUsed = (BYTE)atoi(extractData);												///< ASCII to integer conversion.

			index += strlen(extractData) + COMMA_LENGTH;
			ExtractSubString(sourceString + index, extractData, sizeof(extractData), ',', FALSE);				///< Extracts HDOP value.
			ggaInfoStruct->hdopInfo = (FLOAT)atof(extractData);													///< ASCII to float conversion.

			index += strlen(extractData) + COMMA_LENGTH;
			ExtractSubString(sourceString + index, extractData, sizeof(extractData), ',', TRUE);				///< Extracts altitude.
			GetAltitude(extractData, &(ggaInfoStruct->altitudeInfo));											///< Stores altitude.

			index += strlen(extractData) + COMMA_LENGTH;
			ExtractSubString(sourceString + index, extractData, sizeof(extractData), ',', TRUE);				///< Extracts geoid.
			GetGeoid(extractData, &(ggaInfoStruct->geoidInfo));													///< Stores geoid.

			index += strlen(extractData) + COMMA_LENGTH;
			ExtractSubString(sourceString + index, extractData, sizeof(extractData), ',', FALSE);				///< Extracts time update.
			ggaInfoStruct->timeDgps = (FLOAT)atof(extractData);													///< ASCII to float conversion.

			index += strlen(extractData) + COMMA_LENGTH;
			ExtractSubString(sourceString + index, extractData, sizeof(extractData), '*', FALSE);				///< Extracts DGPS station id.
			ggaInfoStruct->stationIdDgps = atoi(extractData);													///< ASCII to integer conversion.

			index += strlen(extractData) + ASTERISK_LENGTH;
			ExtractSubString(sourceString + index, extractData, sizeof(extractData), '\0', FALSE);				///< Extracts checksum received.
			ConvertAsciiToHex(extractData, (DWORD*)&ggaInfoStruct->checksumData);								///< ASCII to hex conversion.
		
			GenerateCheckSum(sourceString, &checkSum);															///< Generates checksum for sourceString.
			if(ggaInfoStruct->checksumData == checkSum)															///< Checksum verification.
			{
				ggaInfoStruct->checkSumVerify = 'Y';
			}
			else
			{
				ggaInfoStruct->checkSumVerify = 'N';
			}	
			retStatus = TRUE;
			break;
		}
	case GPGLL:
		{
			gllInfoStruct = (gpgllParameter*)ptrStoreResults;

			ExtractSubString(sourceString + index, extractData, sizeof(extractData), ',', TRUE);				///< Extracts latitude.
			GetLatitude(extractData, &(gllInfoStruct->latitudeInfo));											///< Stores latitude.

			index += strlen(extractData) + COMMA_LENGTH;
			ExtractSubString(sourceString + index, extractData, sizeof(extractData), ',', TRUE);				///< Extracts longitude.
			GetLongitude(extractData, &(gllInfoStruct->longitudeInfo));											///< Stores longitude.

			index += strlen(extractData) + COMMA_LENGTH;
			ExtractSubString(sourceString + index, extractData, sizeof(extractData), ',', FALSE);				///< Extracts UTC time.
			GetUtcTime(extractData, &(gllInfoStruct->currentTimeInfo));											///< Stores UTC time.
			
			index += strlen(extractData) + COMMA_LENGTH;
			ExtractSubString(sourceString + index, extractData, sizeof(extractData), ',', FALSE);				///< Extracts validation status of data.
			if(*extractData == 'A' && strlen(extractData) == 1)
			{
				gllInfoStruct->statusOfData = VALID;
			}
			else if(*extractData == 'V' && strlen(extractData) == 1)
			{
				gllInfoStruct->statusOfData = INVALID;
			}
			else
			{
				gllInfoStruct->statusOfData = UNDEFINED_STATUS;
			}

			index += strlen(extractData) + COMMA_LENGTH;
			ExtractSubString(sourceString + index, extractData, sizeof(extractData), '*', FALSE);				///< Extracts position validation.
			gllInfoStruct->positionInfo = extractData[0];

			index += strlen(extractData) + ASTERISK_LENGTH;
			ExtractSubString(sourceString + index, extractData, sizeof(extractData), '\0', FALSE);				///< Extract checksum received.
			ConvertAsciiToHex(extractData, (DWORD*)&gllInfoStruct->checksumData);								///< ASCII to hex conversion.

			GenerateCheckSum(sourceString, &checkSum);															///< Generates checksum for sourceString.
			if(gllInfoStruct->checksumData == checkSum)															///< Checksum verification.
			{
				gllInfoStruct->checkSumVerify = 'Y';
			}
			else
			{
				gllInfoStruct->checkSumVerify = 'N';
			}
			retStatus = TRUE;
			break;
		}
	case GPRMC:
		{			
			rmcInfoStruct = (gprmcParameter*)ptrStoreResults;

			ExtractSubString(sourceString + index, extractData, sizeof(extractData), ',', FALSE);				///< Extracts UTC time.
			GetUtcTime(extractData, &(rmcInfoStruct->currentTimeInfo));

			index += strlen(extractData) + COMMA_LENGTH;
			ExtractSubString(sourceString + index, extractData, sizeof(extractData), ',', FALSE);				///< Extracts validation status of data

			if(*extractData == 'A' && strlen(extractData) == 1)
			{
				rmcInfoStruct->statusOfData = VALID;
			}
			else if(*extractData == 'V' && strlen(extractData) == 1)
			{
				rmcInfoStruct->statusOfData = INVALID;
			}
			else
			{
				rmcInfoStruct->statusOfData = UNDEFINED_STATUS;
			}

			index += strlen(extractData) + COMMA_LENGTH;
			ExtractSubString(sourceString + index, extractData, sizeof(extractData), ',', TRUE);				///< Extracts latitude.
			GetLatitude(extractData, &(rmcInfoStruct->latitudeInfo));											///< Stores latitude.

			index += strlen(extractData) + COMMA_LENGTH;
			ExtractSubString(sourceString + index, extractData, sizeof(extractData), ',', TRUE);				///< Extracts longitude.
			GetLongitude(extractData, &(rmcInfoStruct->longitudeInfo));											///< Stores longitude.

			index += strlen(extractData) + COMMA_LENGTH;
			ExtractSubString(sourceString + index, extractData, sizeof(extractData), ',', FALSE);				///< Extracts speed in Knots.
			rmcInfoStruct->speedInKmph = (FLOAT)atof(extractData) * 1.852f;										///< 1 Knot = 1.852 KMPH.

			index += strlen(extractData) + COMMA_LENGTH;
			ExtractSubString(sourceString + index, extractData, sizeof(extractData), ',', FALSE);				///< Extracts true course over ground.
			rmcInfoStruct->trueCourse = (FLOAT)atof(extractData);												///< ASCII to float conversion.

			index += strlen(extractData) + COMMA_LENGTH;
			ExtractSubString(sourceString + index, extractData, sizeof(extractData), ',', FALSE);				///< Extracts UTC date.
			GetUtcDate(extractData, &(rmcInfoStruct->currentDateInfo));											///< Stores UTC date.

			index += strlen(extractData) + COMMA_LENGTH;
			ExtractSubString(sourceString + index, extractData, sizeof(extractData), ',', TRUE);				///< Extracts magnetic variation.
			GetMagneticVariation(extractData, &(rmcInfoStruct->magneticData));									///< Stores magnetic variation.

			index += strlen(extractData) + COMMA_LENGTH;
			ExtractSubString(sourceString + index, extractData, sizeof(extractData), '*', FALSE);				///< Extracts position validation.
			rmcInfoStruct->positionInfo = extractData[0];														

			index += strlen(extractData) + ASTERISK_LENGTH;
			ExtractSubString(sourceString + index, extractData, sizeof(extractData), '\0', FALSE);				///< Extracts checksum received.
			ConvertAsciiToHex(extractData, (DWORD*)&rmcInfoStruct->checksumData);								///< ASCII to hex conversion.

			GenerateCheckSum(sourceString, &checkSum);
			if(rmcInfoStruct->checksumData == checkSum)
			{
				rmcInfoStruct->checkSumVerify = 'Y';
			}
			else
			{
				rmcInfoStruct->checkSumVerify = 'N';
			}
			retStatus = TRUE;
			break;
		}
	default:
		{
			*error = INVALID_DATATYPE;
		}
	}
	return retStatus;
}