/// @file			gpslib.h
/// @copyright  	Copyright (c) 2014 Toradex AG.
/// $Author:		khv
/// $Date: 			$ 2014-06-04 $
/// @brief:			File to support GPS interfacing.      

#ifndef _GPSLIB_H_
#define _GPSLIB_H_

/// Header files.
#include <windows.h>

/// Macros.
#define MAX_GPS_DATA_STRING_LENGTH	100
#define GPGGA_TOTAL_FIELDS 			 14
#define GPGLL_TOTAL_FIELDS 			  7
#define GPRMC_TOTAL_FIELDS 			 12
#define BITS_IN_NIBBLE 				  4

/// GPS data type.
typedef enum tGpsDataType
{
	GPGGA,						///< Global positioning system fixed data.  		
	GPGLL,						///< Geographic latitude- longitude (GLL) position.
	GPRMC						///< Recommended minimum specific (RMC) GPS data.
}gpsDataType;

/// Position parameter.
typedef enum tPositionparameter
{
	FAILURE,					///< Invalid- receiver cannot report a position fix.
	VALID_SPS,					///< GPS position fix using Standard Position Service.
	VALID_DGPS,					///< GPS position fix using Differential Position Service.
	VALID_PPS,					///< GPS position fix using Precise Positioning Service.
	REAL_TIME_KINEMATIC,
	FLOAT_RTK,
	ESTIMATED,
	MANUAL_INPUT_MODE,
	SIMULATION_MODE,
	INVALID_POSITION
}positionParameter;

/// Status parameter.
typedef enum tStatusParameter
{
	VALID,
	INVALID,
	UNDEFINED_STATUS
}statusParameter;

/// Type of error.
typedef enum tErrorType
{
	ERROR_FREE,
	COM_PORT_READ_FAILED,
	DATATYPE_NOT_FOUND,
	INVALID_DATATYPE,
	INVALID_DATA,
	INVALID_FIELD,
	CHECKSUM_INCORRECT
}errorType;

/// UTC time in Hour, Minute, and Second.
typedef struct tTimeParameters
{
	BYTE utcHour;
	BYTE utcMin;
	FLOAT utcSec;
}timeParameter;

/// UTC date in Date, Month, and year.
typedef struct tDateParameters
{
	BYTE utcDate;
	BYTE utcMonth;
	BYTE utcYear;
}dateParameter;

/// Latitude in Degree, Minute, and Second.
typedef struct tLatitudeParameters
{
	BYTE latDeg;
	BYTE latMin;
	FLOAT latSec;
	CHAR latDirection;
}latitudeParameter;

/// Longitude in Degree, Minute, and Second.
typedef struct tLongitudeParameters
{
	WORD longDeg;
	BYTE longMin;
	FLOAT longSec;
	CHAR longDirection;
}longitudeParameter;

/// Altitude parameter.
typedef struct tAltitudeParameters
{
	FLOAT height;
	CHAR altiUnit; 
}altitudeParameter;

/// Geoid parameter.
typedef struct tGeoidParameters
{
	FLOAT height;
	CHAR geoidUnit;
}geoidParameter;

/// Magnetic variation parameter.
typedef struct tMagnetParameters
{
	FLOAT magVarDeg;
	CHAR magVarDirection;
}magnetParameter;

/// Global positioning system fixed data (Example: $GPGGA,133829.000,1256.1520,N,07737.0638,E,2,5,1.26,912.8,M,-88.4,M,0000,0000*79).
typedef struct tGPSParametersInGGA
{
	errorType error;
	timeParameter currentTimeInfo;
	latitudeParameter latitudeInfo;
	longitudeParameter longitudeInfo;
	positionParameter positionInfo;
	BYTE satelliteUsed;
	FLOAT hdopInfo;
	altitudeParameter altitudeInfo;
	geoidParameter geoidInfo;
	FLOAT timeDgps;
	DWORD stationIdDgps;
	BYTE checksumData;
	CHAR checkSumVerify;
}gpggaParameter;

/// Geographic latitude- longitude (GLL) position. (Example: $GPGLL,3723.2475,N,12158.3416,W,161229.487,A,A*41).
typedef struct tGPSParametersInGLL
{
	errorType error;
	timeParameter currentTimeInfo;
	latitudeParameter latitudeInfo;
	longitudeParameter longitudeInfo;
	statusParameter statusOfData;
	positionParameter positionInfo;
	BYTE checksumData;
	CHAR checkSumVerify;
}gpgllParameter;

/// Recommended minimum specific (RMC) GPS data. (Example: $GPRMC,133830.000,A,1256.1518,N,07737.0636,E,0.93,208.67,030614,003.1,W,D*11).
typedef struct tGPSParametersInRMC
{
	errorType error;
	timeParameter currentTimeInfo;
	dateParameter currentDateInfo;
	latitudeParameter latitudeInfo;
	longitudeParameter longitudeInfo;
	statusParameter statusOfData;
	CHAR positionInfo;
	FLOAT speedInKmph;
	FLOAT trueCourse;
	magnetParameter magneticData;
	BYTE checksumData;
	CHAR checkSumVerify;
}gprmcParameter;

//******************************************************************************
/// Converts ASCII string to hex number (Function results a maximum of 8 hex digits in a DWORD, first ASCII character at most significant nibble of DWORD).
/// @param[in]    source      ASCII string.
/// @param[out]   hexResult	  Store hex value.
/// @retval       TRUE		  SUCCESS.
/// 			  FALSE       Incorrect argument passed to the function.
BOOL ConvertAsciiToHex(CHAR *source, DWORD *hexResult);

//******************************************************************************
/// Generates checksum for source string. (Format = $GPxxx,...*xx)
/// @param[in]    source      Source string.
/// @param[out]   checkSum    Store checksum.
/// @retval       TRUE		  SUCCESS.
/// 			  FALSE       Incorrect argument passed to the function.
BOOL GenerateCheckSum(CHAR *source, BYTE* checkSum);

//******************************************************************************
/// Extracts time parameters from time format string. (Format = HHMMSS[dot]SS..)
/// @param[in]     timeSource  Time in string format.
/// @param[out]    timeStore   Pointer to time structure.
/// @retval        TRUE		   SUCCESS.
/// 			   FALSE       Incorrect argument passed to the function.
BOOL GetUtcTime(CHAR *timeSource, timeParameter *timeStore);

//******************************************************************************
/// Extracts date parameters from date format string. (Format = DDMMYY)
/// @param[in]    dateSource  Date in string format.
/// @param[out]   dateStore   Pointer to date structure.
/// @retval       TRUE		  SUCCESS.
/// 			  FALSE       Incorrect argument passed to the function.
BOOL GetUtcDate(CHAR *dateSource, dateParameter *dateStore);

//******************************************************************************
/// Extracts latitude parameters from latitude format string. (Format = DDMM[dot]MM..,Direction)
/// @param[in]    latitudeSource  Latitude in string format.
/// @param[out]   latitudeStore   Pointer to latitude structure.
/// @retval       TRUE			  SUCCESS.
/// 			  FALSE           Incorrect argument passed to the function.
BOOL GetLatitude(CHAR *latitudeSource, latitudeParameter *latitudeStore);

//******************************************************************************
/// Extracts longitude parameters from longitude format string. (Format = DDMM[dot]MM..,Direction)
/// @param[in]    longitudeSource  Longitude in string format.
/// @param[out]   longitudeStore   Pointer to longitude structure.
/// @retval       TRUE			   SUCCESS.
/// 			  FALSE            Incorrect argument passed to the function.
BOOL GetLongitude(CHAR *longitudeSource, longitudeParameter *longitudeStore);

//******************************************************************************
/// Extracts altitude parameters from altitude format string. (Format = xxx[dot]xx..,Direction)
/// @param[in]    altitudeSource  Altitude in string format.
/// @param[out]   altitudeStore   Pointer to altitude structure.
/// @retval       TRUE			  SUCCESS.
/// 			  FALSE			  Incorrect argument passed to the function.
BOOL GetAltitude(CHAR *altitudeSource, altitudeParameter *altitudeStore);

//******************************************************************************
/// Extracts geoid parameters from geoid format string. (Format = xxx[dot]xx..,Direction)
/// @param[in]    geoidSource    Geoid in string format.
/// @param[out]   geoidStore     Pointer to geoid structure.
/// @retval       TRUE		     SUCCESS.
/// 			  FALSE          Incorrect argument passed to the function.
BOOL GetGeoid(CHAR *geoidSource, geoidParameter *geoidStore);

//******************************************************************************
/// Extracts magnetic parameters from magnetic format string. (Format = xxx[dot]xx..,Direction)
/// @param[in]    magnetSource   Magnetic info in string format.
/// @param[out]   magnetStore    Pointer to magnetic info structure.
/// @retval       TRUE		     SUCCESS.
/// 			  FALSE          Incorrect argument passed to the function.
BOOL GetMagneticVariation(CHAR *magnetSource, magnetParameter *magnetStore);

//******************************************************************************
/// Extracts data from source string up to termination character and store in destination string.
/// @param[in]    sourceString         Source string.
/// @param[out]   destinationString    Destination string.
/// @param[in]    destinationSize      Size of Destination buffer.
/// @param[in]    terminator           Termination character.
/// @param[in]    skipTerminatorOnce   if TRUE, function will skip first occurrence of terminator character and extract up to second occurrence.
/// @retval       TRUE				   SUCCESS.
/// 			  FALSE				   Incorrect argument passed to the function.
BOOL ExtractSubString(CHAR *sourceString, CHAR *destinationString, DWORD destinationSize, CHAR terminator, BOOL skipTerminatorOnce);

//******************************************************************************
/// Extracts individual data field from GPS string and store. (GPS string Format = $GPxxx,.....,*xx)
/// @param[in]    sourceString       String given to extract information from.
/// @param[in]    dataId             GPS data type.
/// @param[out]   ptrStoreResults    Pointer to data storage structure.
/// @param[out]   error              Type of error occurred during processing.
/// @retval       TRUE               Success.
///               FALSE              Failure.
BOOL ExtractAllFieldsFromGpsData(CHAR *srcString, gpsDataType dataId, VOID *ptrStoreResults, errorType *error);
#endif